/*-
 * Copyright (C)2022 @BABOLO http://www.babolo.ru/
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

/* Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#ident "@(#) Copyright (C)2022 @BABOLO http://www.babolo.ru/"
#ident "@(#) $Id: mod_authn_sqlite3.c,v 1.2 2022/05/21 23:38:41 babolo Exp $"

#include <apr_strings.h>
#include <apr_md5.h>
#include <ap_config.h>
#include <ap_provider.h>
#include <httpd.h>
#include <sqlite3.h>
#include <http_config.h>
#include <http_core.h>
#include <http_log.h>
#include <http_protocol.h>
#include <http_request.h>
#include <mod_auth.h>

#define MAXPWSZ 100

typedef struct {
    char *dbfile;
    char *passwd;
    char *digest;
    char *context;
    char *allpass;
    char  pop[];
} authn_sqlite3_config_rec;

module AP_MODULE_DECLARE_DATA authn_sqlite3_module;

const static char f[] =
    "SELECT cred\n"
    " FROM \"%s\"\n"
    " WHERE login = ? AND (passwd = ? OR passwd IS NULL)\n"
    "ORDER BY cred\n"
    "LIMIT 1\n"
    ";";

const static char c[] =
    "SELECT cred\n"
    " FROM \"%s\"\n"
    " WHERE login = ? AND (passwd = ? OR passwd IS NULL) AND context = ?\n"
    "ORDER BY cred\n"
    "LIMIT 1\n"
    ";";

const static char u[] =
    " *+,-./0123456789:=@ABCDEFGHIJKLMNOPQRSTUVWXYZ_abcdefghijklmnopqrstuvwxyz";

static const command_rec authn_sqlite3_cmds[] =
{   AP_INIT_TAKE1( "Auth_SQLite3_File"
                 , ap_set_file_slot
                 , (void *)APR_OFFSETOF(authn_sqlite3_config_rec, dbfile)
                 , OR_AUTHCFG
                 , "SQLite3 DB file containing user IDs and passwords"
                 )
,   AP_INIT_TAKE1( "Auth_SQLite3_Passwd"
                 , ap_set_string_slot
                 , (void *)APR_OFFSETOF(authn_sqlite3_config_rec, passwd)
                 , OR_AUTHCFG
                 , "Name of passwd table"
                 )
,   AP_INIT_TAKE1( "Auth_SQLite3_Digest"
                 , ap_set_string_slot
                 , (void *)APR_OFFSETOF(authn_sqlite3_config_rec, digest)
                 , OR_AUTHCFG
                 , "Name of digest table"
                 )
,   AP_INIT_TAKE1( "Auth_SQLite3_Context"
                 , ap_set_string_slot
                 , (void *)APR_OFFSETOF(authn_sqlite3_config_rec, context)
                 , OR_AUTHCFG
                 , "Context of AAA"
                 )
,   AP_INIT_TAKE1( "Auth_SQLite3_AllPass"
                 , ap_set_string_slot
                 , (void *)APR_OFFSETOF(authn_sqlite3_config_rec, allpass)
                 , OR_AUTHCFG
                 , "All pass URI begin"
                 )
,   {NULL}
};

static void
testset(request_rec *r) {
    if  (!!r->connection->client_ip)
        apr_table_setn(r->subprocess_env, "TEST_CIP"  , r->connection->client_ip  );
    if  (!!r->connection->remote_host)
        apr_table_setn(r->subprocess_env, "TEST_RHOST", r->connection->remote_host);
    if  (!!r->connection->local_ip)
        apr_table_setn(r->subprocess_env, "TEST_LIP"  , r->connection->local_ip   );
    if  (!!r->connection->local_host)
        apr_table_setn(r->subprocess_env, "TEST_LHOST", r->connection->local_host );
    if  (!!r->uri)            apr_table_setn(r->subprocess_env, "TEST_URI"  , r->uri           );
    if  (!!r->args)           apr_table_setn(r->subprocess_env, "TEST_ARGS" , r->args          );
    if  (!!r->protocol)       apr_table_setn(r->subprocess_env, "TEST_PROTO", r->protocol      );
    if  (!!r->the_request)    apr_table_setn(r->subprocess_env, "TEST_REQ"  , r->the_request   );
    if  (!!r->hostname)       apr_table_setn(r->subprocess_env, "TEST_HOST" , r->hostname      );
    if  (!!r->method)         apr_table_setn(r->subprocess_env, "TEST_METH" , r->method        );
    if  (!!r->user)           apr_table_setn(r->subprocess_env, "TEST_USER" , r->user          );
    if  (!!r->unparsed_uri)   apr_table_setn(r->subprocess_env, "TEST_UURI" , r->unparsed_uri  );
    if  (!!r->path_info)      apr_table_setn(r->subprocess_env, "TEST_PATH" , r->path_info     );
    if  (!!r->useragent_ip)   apr_table_setn(r->subprocess_env, "TEST_UIP"  , r->useragent_ip  );
    if  (!!r->useragent_host) apr_table_setn(r->subprocess_env, "TEST_UHOST", r->useragent_host);
}

static void *
create_authn_sqlite3_dir_config(apr_pool_t *p, char *d) {
    authn_sqlite3_config_rec *conf;

    conf = apr_palloc(p, sizeof(authn_sqlite3_config_rec) + strlen(c) + MAXPWSZ);
    bzero(conf, sizeof(authn_sqlite3_config_rec) + strlen(c));
    return(conf);
}

static authn_status
check_password(request_rec *r, const char *user, const char *password) {
    authn_sqlite3_config_rec *conf   = ap_get_module_config(r->per_dir_config, &authn_sqlite3_module);
    sqlite3                  *odf;
    sqlite3_stmt             *vm;
    int                       ex;
    int                       i;

    if  (!conf->dbfile) {
        ap_log_rerror( APLOG_MARK
                     , APLOG_ERR
                     , 0
                     , r
                     , "mod_authn_sqlite3@BABOLO Auth_SQLite3_File not specified"
                     );
        return(AUTH_GENERAL_ERROR);
    }
    if  (!conf->passwd) {
        ap_log_rerror( APLOG_MARK
                     , APLOG_ERR
                     , 0
                     , r
                     , "mod_authn_sqlite3@BABOLO Auth_SQLite3_Passwd not specified"
                     );
        return(AUTH_GENERAL_ERROR);
    }
    if  (!conf->pop[0]) {
        size_t  sz = 0;

        if  (!!conf->passwd[strspn(conf->passwd, u)]) {
            ap_log_rerror( APLOG_MARK
                         , APLOG_ERR
                         , 0
                         , r
                         , "mod_authn_sqlite3@BABOLO Illegal table name: %s"
                         , conf->passwd
                         );
            return(AUTH_GENERAL_ERROR);
        }
        sz = strlen((!!conf->context) ? c : f) + strlen(conf->passwd);
        if  (sz > strlen(c) + MAXPWSZ) {
            ap_log_rerror( APLOG_MARK
                         , APLOG_ERR
                         , 0
                         , r
                         , "mod_authn_sqlite3@BABOLO Too big tablename %s"
                         , conf->passwd
                         );
            return(AUTH_GENERAL_ERROR);
        }
        snprintf(conf->pop, sz + 1, (!!conf->context) ? c : f, conf->passwd);
    }
    if  (!!(ex = sqlite3_initialize())) {
        ap_log_rerror( APLOG_MARK
                     , APLOG_ERR
                     , 0
                     , r
                     , "mod_authn_sqlite3@BABOLO Could not initialize %d"
                     , ex
                     );
        return(AUTH_GENERAL_ERROR);
    }
    if  (!!(ex = sqlite3_open_v2(conf->dbfile, &odf, SQLITE_OPEN_READONLY, NULL))) {
        sqlite3_shutdown();
        ap_log_rerror( APLOG_MARK
                     , APLOG_ERR
                     , 0
                     , r
                     , "mod_authn_sqlite3@BABOLO Could not open password DB %d: %s"
                     , ex
                     , conf->dbfile
                     );
        return(AUTH_GENERAL_ERROR);
    }
    if  (!!(ex = sqlite3_prepare_v2(odf, conf->pop, -1, &vm, NULL))) {
        sqlite3_close(odf);
        ap_log_rerror( APLOG_MARK
                     , APLOG_ERR
                     , 0
                     , r
                     , "mod_authn_sqlite3@BABOLO Could not compile %d: %s"
                     , ex
                     , conf->pop
                     );
        if  (!vm || !!sqlite3_finalize(vm) || !!sqlite3_shutdown()) {
            ap_log_rerror( APLOG_MARK
                         , APLOG_ERR
                         , 0
                         , r
                         , "mod_authn_sqlite3@BABOLO Could not complete"
                         );
        }
        return(AUTH_GENERAL_ERROR);
    }
    i = 0;
    if  (!!conf->context && !!(ex = sqlite3_bind_text(vm, ++i, conf->context, -1, SQLITE_STATIC))) {
        sqlite3_finalize(vm);
        sqlite3_close(odf);
        sqlite3_shutdown();
        ap_log_rerror( APLOG_MARK
                     , APLOG_ERR
                     , 0
                     , r
                     , "mod_authn_sqlite3@BABOLO Could not bind[%d] %d: %s"
                     , i
                     , ex
                     , conf->context
                     );
        return(AUTH_GENERAL_ERROR);
    }
    if  (!!(ex = sqlite3_bind_text(vm, ++i, user, -1, SQLITE_STATIC))) {
        sqlite3_finalize(vm);
        sqlite3_close(odf);
        sqlite3_shutdown();
        ap_log_rerror( APLOG_MARK
                     , APLOG_ERR
                     , 0
                     , r
                     , "mod_authn_sqlite3@BABOLO Could not bind[%d] %d: %s"
                     , i
                     , ex
                     , user
                     );
        return(AUTH_GENERAL_ERROR);
    }
    if  (!!(ex = sqlite3_bind_text(vm, ++i, password, -1, SQLITE_STATIC))) {
        sqlite3_finalize(vm);
        sqlite3_close(odf);
        sqlite3_shutdown();
        ap_log_rerror( APLOG_MARK
                     , APLOG_ERR
                     , 0
                     , r
                     , "mod_authn_sqlite3@BABOLO Could not bind[%d] %d: %s"
                     , i
                     , ex
                     , password
                     );
        return(AUTH_GENERAL_ERROR);
    }
    switch((ex = sqlite3_step(vm))) {
    case SQLITE_ROW: {
            const u_char *tuple = NULL;
            char         *cred;
            int           ncol;

            ncol = sqlite3_column_count(vm);
            if  (ncol != 1) {
                sqlite3_finalize(vm);
                sqlite3_close(odf);
                sqlite3_shutdown();
                ap_log_rerror( APLOG_MARK
                             , APLOG_ERR
                             , 0
                             , r
                             , "mod_authn_sqlite3@BABOLO Col num %d IN %s"
                             , ncol
                             , conf->pop
                             );
                return(AUTH_GENERAL_ERROR);
            } else {
                if  (sqlite3_column_type(vm, 0) != SQLITE_NULL) tuple = sqlite3_column_text(vm, 0);
                if  (!!tuple) {
                    cred = apr_pstrdup(r->pool, (const char *)tuple);
                    apr_table_setn(r->subprocess_env, "CREDENTIALS", cred);
                }
                testset(r);
                sqlite3_finalize(vm);
                sqlite3_close(odf);
                sqlite3_shutdown();
                return(AUTH_GRANTED);
        }   }
    case SQLITE_DONE:
        sqlite3_finalize(vm);
        sqlite3_close(odf);
        sqlite3_shutdown();
        if  (!!conf->allpass && !strcmp(r->uri, conf->allpass)) {
            apr_table_setn(r->subprocess_env, "CREDENTIALS", "0");
            testset(r);
            return(AUTH_GRANTED);
        }
        return(AUTH_DENIED);
    }
    sqlite3_finalize(vm);
    sqlite3_close(odf);
    sqlite3_shutdown();
    ap_log_rerror( APLOG_MARK
                 , APLOG_ERR
                 , 0
                 , r
                 , "mod_authn_sqlite3@BABOLO Could not SELECT step %d"
                 , ex
                 );
    return(AUTH_GENERAL_ERROR);
}

static authn_status
get_realm_hash(request_rec *r, const char *user, const char *realm, char **rethash) {
    char                     *sqlite3_hash = NULL;
    authn_sqlite3_config_rec *conf = ap_get_module_config(r->per_dir_config, &authn_sqlite3_module);

    if  (!conf->dbfile) {
        ap_log_rerror( APLOG_MARK
                     , APLOG_ERR
                     , 0
                     , r
                     , "mod_authn_sqlite3@BABOLO Auth_SQLite3_File not specified"
                     );
        return AUTH_GENERAL_ERROR;
    }




    if (!sqlite3_hash) return(AUTH_USER_NOT_FOUND);
    *rethash = sqlite3_hash;
    return(AUTH_USER_FOUND);
}

static const authn_provider authn_sqlite3_provider =
{   &check_password
,   &get_realm_hash
};

static void
register_hooks(apr_pool_t *p) {
    ap_register_auth_provider( p
                             , AUTHN_PROVIDER_GROUP
                             , "sqlite3"
                             , AUTHN_PROVIDER_VERSION
                             , &authn_sqlite3_provider
                             , AP_AUTH_INTERNAL_PER_CONF
                             );
}

AP_DECLARE_MODULE(authn_sqlite3) =
{   STANDARD20_MODULE_STUFF
,   create_authn_sqlite3_dir_config /* dir config creater */
,   NULL                            /* dir merger --- default is to override */
,   NULL                            /* server config */
,   NULL                            /* merge server config */
,   authn_sqlite3_cmds              /* command apr_table_t */
,   register_hooks                  /* register hooks */
};
